@@ -3,18 +3,18 @@
 const Clutter = imports.gi.Clutter;
 const Gio = imports.gi.Gio;
 const GLib = imports.gi.GLib;
-const Lang = imports.lang;
+const GObject = imports.gi.GObject;
+const Pango = imports.gi.Pango;
 const Mainloop = imports.mainloop;
 const Meta = imports.gi.Meta;
 const St = imports.gi.St;
 const Cinnamon = imports.gi.Cinnamon;
-const Signals = imports.signals;
 
 const FileUtils = imports.misc.fileUtils;
 const Main = imports.ui.main;
+const Dialog = imports.ui.dialog;
 const ModalDialog = imports.ui.modalDialog;
 const CinnamonEntry = imports.ui.cinnamonEntry;
-const Tweener = imports.ui.tweener;
 const Util = imports.misc.util;
 const History = imports.misc.history;
 
@@ -32,22 +32,32 @@ const EXEC_ARG_KEY = 'exec-arg';
 const SHOW_COMPLETIONS_KEY = 'run-dialog-show-completions';
 const ALIASES_KEY = 'run-dialog-aliases';
 
-const DIALOG_GROW_TIME = 0.1;
 const MAX_COMPLETIONS = 40;
 
+const NAVIGATE_TYPE_NONE = 0;
+const NAVIGATE_TYPE_TAB = 1;
+const NAVIGATE_TYPE_ARROW = 2;
+
+const UP = 1;
+const DOWN = 2;
+
 const DEVEL_COMMANDS = { 'lg': x => Main.createLookingGlass().open(),
-                         'r': x => global.reexec_self(),
-                         'restart': x => global.reexec_self(),
+                         'r': x => Main.restartCinnamon(true),
+                         'restart': x => Main.restartCinnamon(true),
                          'debugexit': x => Meta.quit(Meta.ExitCode.ERROR),
                          'rt': x => Main.themeManager._changeTheme() };
 
+/* The modal dialog parent class has a 100ms close animation.  Delay long enough for it
+ * to complete before doing something disruptive like restarting cinnamon */
+const DEVEL_COMMAND_DELAY =  parseInt(ModalDialog.OPEN_AND_CLOSE_TIME) + 10;
+
 /**
  * completeCommand:
  * @text (string): initial string to complete.
  *
  * This function finds possible command completions for @text. @text is first
- * split at whitspaces, and completion is performed on the last segment. Note
- * that this currently does not recognize escaped whitspaces.
+ * split at whitespaces, and completion is performed on the last segment. Note
+ * that this currently does not recognize escaped whitespaces.
  *
  * If the last segment starts with a `/`, then it is considered to be an
  * absolute path. Otherwise, if it is the first segment (ie. there is only one
@@ -134,73 +144,80 @@ function completeCommand(text) {
     return [common.substring(last.length, common.length), results.map(x => x.substring(last.length, x.length))];
 }
 
-function RunDialog() {
-    this._init();
-}
-
-RunDialog.prototype = {
-__proto__: ModalDialog.ModalDialog.prototype,
-    _init : function() {
-        ModalDialog.ModalDialog.prototype._init.call(this, { styleClass: 'run-dialog' });
+var RunDialog = GObject.registerClass(
+class RunDialog extends ModalDialog.ModalDialog {
+    _init() {
+        super._init({ styleClass: 'run-dialog'});
 
         this._lockdownSettings = new Gio.Settings({ schema_id: LOCKDOWN_SCHEMA });
         this._terminalSettings = new Gio.Settings({ schema_id: TERMINAL_SCHEMA });
-        global.settings.connect('changed::development-tools', Lang.bind(this, function () {
+        global.settings.connect('changed::development-tools', () =>  {
             this._enableInternalCommands = global.settings.get_boolean('development-tools');
-        }));
+        });
         this._enableInternalCommands = global.settings.get_boolean('development-tools');
 
         global.display.connect('restart', () => this.close());
 
-        let label = new St.Label({ style_class: 'run-dialog-label',
-                                   text: _("Please enter a command:") });
+        let title = _("Run a Command");
 
-        this.contentLayout.add(label, { y_align: St.Align.START });
+        let content = new Dialog.MessageDialogContent({ title });
+        this.contentLayout.add_actor(content);
 
         let entry = new St.Entry({ style_class: 'run-dialog-entry' });
         CinnamonEntry.addContextMenu(entry);
 
-        entry.label_actor = label;
-
         this._entryText = entry.clutter_text;
         this._oldText = "";
-        this.contentLayout.add(entry, { y_align: St.Align.START });
+        content.add_child(entry);
         this.setInitialKeyFocus(this._entryText);
 
         this._completionBox = new St.Label({style_class: 'run-dialog-completion-box'});
-        this.contentLayout.add(this._completionBox);
+        content.add_child(this._completionBox);
         this._completionSelected = 0;
 
-        this._errorBox = new St.BoxLayout({ style_class: 'run-dialog-error-box' });
+        let defaultDescriptionText = _("Press ESC to close");
 
-        this.contentLayout.add(this._errorBox, { expand: true });
-
-        let errorIcon = new St.Icon({ icon_name: 'dialog-error', icon_size: 24, style_class: 'run-dialog-error-icon' });
-
-        this._errorBox.add(errorIcon, { y_align: St.Align.MIDDLE });
+        this._descriptionLabel = new St.Label({
+            style_class: 'run-dialog-description',
+            text: defaultDescriptionText
+        });
+        this._descriptionLabel.clutter_text.line_wrap = true;
+        this._descriptionLabel.clutter_text.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR);
+        content.add_child(this._descriptionLabel);
 
         this._commandError = false;
 
-        this._errorMessage = new St.Label({ style_class: 'run-dialog-error-label' });
-        this._errorMessage.clutter_text.line_wrap = true;
+        this._entryText.connect('key-press-event', this._onKeyPress.bind(this));
 
-        this._errorBox.add(this._errorMessage, { expand: true,
-                                                 y_align: St.Align.MIDDLE,
-                                                 y_fill: false });
-
-        this._errorBox.hide();
-
-        this._history = new History.HistoryManager({ gsettingsKey: HISTORY_KEY,
-                                                     entry: this._entryText,
-                                                     deduplicate: true });
-        this._entryText.connect('key-press-event', Lang.bind(this, this._onKeyPress));
+        this._history = new History.HistoryManager({
+            gsettingsKey: HISTORY_KEY,
+            entry: this._entryText,
+            deduplicate: true
+        });
 
         this._updateCompletionTimer = 0;
-     },
+     }
 
-    _onKeyPress: function (o, e) {
+    _onKeyPress(o, e) {
         let symbol = e.get_key_symbol();
-        if (symbol == Clutter.Return || symbol == Clutter.KP_Enter) {
+        if (symbol === Clutter.KEY_Return || symbol === Clutter.KEY_KP_Enter) {
+            if (o.get_text().trim() == "") {
+                return false;
+            }
+
+            /* When enter is hit with completions open, if the current selection
+             * is a folder, open that folder immediately.  Otherwise, just close
+             * the completion box - the user can add an argument to the command
+             * they selected (there's already a space provided) */
+            if (this._completionBox.visible && !o.get_text().endsWith("/")) {
+                this._completionSelected = 0;
+                this._completionBox.hide();
+                this._entryText.set_selection_bound(-1);
+                this._entryText.set_cursor_position(-1);
+                this._oldText = "";
+                return true;
+            }
+
             this.popModal();
             if (Cinnamon.get_event_state(e) & Clutter.ModifierType.CONTROL_MASK)
                 this._run(o.get_text(), true);
@@ -214,15 +231,26 @@ __proto__: ModalDialog.ModalDialog.prototype,
             }
             return true;
         }
-        if (symbol == Clutter.Escape || symbol == Clutter.Super_L || symbol == Clutter.Super_R) {
+        if (symbol === Clutter.KEY_Escape || symbol === Clutter.KEY_Super_L || symbol === Clutter.KEY_Super_R) {
             this.close();
             return true;
         }
-        if (symbol == Clutter.Tab) {
-            this._updateCompletions(true);
+        if (symbol === Clutter.KEY_Tab) {
+            this._updateCompletions(NAVIGATE_TYPE_TAB);
             return true;
         }
-        if (symbol == Clutter.BackSpace) {
+
+        if (this._completionBox.visible) {
+            if (symbol === Clutter.KEY_Up) {
+                this._updateCompletions(NAVIGATE_TYPE_ARROW, UP);
+                return true;
+            } else if (symbol === Clutter.KEY_Down) {
+                this._updateCompletions(NAVIGATE_TYPE_ARROW, DOWN);
+                return true;
+            }
+        }
+
+        if (symbol === Clutter.KEY_BackSpace) {
             this._completionSelected = 0;
             this._completionBox.hide();
             this._oldText = "";
@@ -234,18 +262,18 @@ __proto__: ModalDialog.ModalDialog.prototype,
                 this._updateCompletionTimer = 0;
             }
 
-            this._updateCompletionTimer = Mainloop.timeout_add(200, Lang.bind(this, this._updateCompletions));
+            this._updateCompletionTimer = Mainloop.timeout_add(200, this._updateCompletions.bind(this));
             return false;
         }
         return false;
-    },
+    }
 
     // There is different behaviour depending on whether this is called due to
     // pressing tab or other keys.
-    _updateCompletions: function(tab) {
+    _updateCompletions(nav_type=NAVIGATE_TYPE_NONE, direction=DOWN) {
         this._updateCompletionTimer = 0;
 
-        let text = this._entryText.get_text();
+        let text = this._expandHome(this._entryText.get_text());
 
         /* If update is caused by user pressing key, and the user just finished
          * a directory path, don't provide new predictions. For example, the
@@ -256,24 +284,34 @@ __proto__: ModalDialog.ModalDialog.prototype,
          * not perform completions, since completions will list all files in
          * /home/user/, which is unexpected.
          */
-        if (!tab && text.charAt(text.length - 1) == "/") {
+        if (!nav_type && text.charAt(text.length - 1) == "/") {
             this._completionBox.hide();
             this._oldText = "";
             return;
         }
 
-        // Currnet suggested completion is selected. Do not include in query.
+        // Current suggested completion is selected. Do not include in query.
         text = text.slice(0, text.lastIndexOf(this._entryText.get_selection()));
 
         /* If update is caused by user typing "tab" and no text has changed
          * since then, cycle through available completions.
          */
-        if (this._oldText == text && tab && this._completionBox.visible) {
-            this._completionSelected ++;
-            this._completionSelected %= this._completions.length;
-            this._showCompletions(text);
-            return;
+        if (this._oldText == text && nav_type && this._completionBox.visible) {
+            if ((nav_type == NAVIGATE_TYPE_ARROW && direction == DOWN) || nav_type == NAVIGATE_TYPE_TAB) {
+                this._completionSelected ++;
+                this._completionSelected %= this._completions.length;
+                this._showCompletions(text);
+                return;
+            } else { // nav_type was > 0 and not tab, and not down, so navigate UP.
+                if (this._completionSelected > 0) {
+                    this._completionSelected --;
+                }
+
+                this._showCompletions(text);
+                return;
+            }
         }
+
         this._oldText = text;
 
         let [postfix, completions] = completeCommand(text);
@@ -286,7 +324,7 @@ __proto__: ModalDialog.ModalDialog.prototype,
          * possible completions. If there is no possible completion, then hide
          * completion box.
          */
-        if (postfix.length > 0 && tab) {
+        if (postfix.length > 0 && nav_type == NAVIGATE_TYPE_TAB) {
             this._entryText.set_text(text + postfix);
         } else if (completions.length > 0 &&
                 global.settings.get_boolean(SHOW_COMPLETIONS_KEY)) {
@@ -297,9 +335,18 @@ __proto__: ModalDialog.ModalDialog.prototype,
             this._completionBox.hide();
             this._oldText = "";
         }
-    },
+    }
 
-    _showCompletions: function(orig) {
+    _expandHome(text) {
+        if (text.charAt(0) == '~') {
+            text = text.slice(1);
+            return GLib.build_filenamev([GLib.get_home_dir(), text]);
+        }
+
+        return text;
+    }
+
+    _showCompletions(orig) {
         /* Show a list of possible completions, and allow users to scroll
          * through them. The scrolling mechanism is done in _updateCompletions,
          * which provides the current selected index in
@@ -339,16 +386,19 @@ __proto__: ModalDialog.ModalDialog.prototype,
 
         this._completionBox.clutter_text.set_markup(text);
         this._entryText.set_selection(-1, orig.length);
-    },
+    }
 
-    _run : function(input, inTerminal) {
+    _run(input, inTerminal) {
+        input = input.trim();
         this._history.addItem(input);
         this._commandError = false;
-        if (this._enableInternalCommands && input.trim() in DEVEL_COMMANDS) {
-            DEVEL_COMMANDS[input.trim()]();
+        if (this._enableInternalCommands && input in DEVEL_COMMANDS) {
+            Mainloop.timeout_add(DEVEL_COMMAND_DELAY, ()=>DEVEL_COMMANDS[input]());
             return;
         }
 
+        input = this._expandHome(input);
+
         // Aliases is a list of strings of the form a:b, where an instance of
         // "a" is to be replaced with "b". Replacement is only performed on the
         // first word
@@ -365,23 +415,21 @@ __proto__: ModalDialog.ModalDialog.prototype,
         try {
             if (inTerminal) {
                 let exec = this._terminalSettings.get_string(EXEC_KEY);
-                let exec_arg = this._terminalSettings.get_string(EXEC_ARG_KEY);
-                command = exec + ' ' + exec_arg + ' ' + input;
+                let execArg = this._terminalSettings.get_string(EXEC_ARG_KEY);
+                command = exec + ' ' + execArg + ' ' + input;
             }
-            Util.trySpawnCommandLine(command);
+            Util.spawnCommandLineAsync(command, null, null);
         } catch (e) {
             // Mmmh, that failed - see if @input matches an existing file
             let path = null;
-            input = input.trim();
+
             if (input.charAt(0) == '/') {
                 path = input;
             } else {
-                if (input.charAt(0) == '~')
-                    input = input.slice(1);
                 path = GLib.build_filenamev([GLib.get_home_dir(), input]);
             }
 
-            if (GLib.file_test(path, GLib.FileTest.EXISTS)) {
+            if (path && GLib.file_test(path, GLib.FileTest.EXISTS)) {
                 let file = Gio.file_new_for_path(path);
                 try {
                     Gio.app_info_launch_default_for_uri(file.get_uri(),
@@ -398,33 +446,19 @@ __proto__: ModalDialog.ModalDialog.prototype,
                 this._showError(e.message);
             }
         }
-    },
+    }
 
-    _showError : function(message) {
+    _showError(message) {
         this._commandError = true;
 
-        this._errorMessage.set_text(message);
+        this._descriptionLabel.set_text(message.trim());
+        this._descriptionLabel.add_style_class_name('error');
+    }
 
-        if (!this._errorBox.visible) {
-            let [errorBoxMinHeight, errorBoxNaturalHeight] = this._errorBox.get_preferred_height(-1);
-
-            let parentActor = this._errorBox.get_parent();
-            Tweener.addTween(parentActor,
-                             { height: parentActor.height + errorBoxNaturalHeight,
-                               time: DIALOG_GROW_TIME,
-                               transition: 'easeOutQuad',
-                               onComplete: Lang.bind(this,
-                                                     function() {
-                                                         parentActor.set_height(-1);
-                                                         this._errorBox.show();
-                                                     })
-                             });
-        }
-    },
-
-    open: function() {
+    open() {
         this._history.lastItem();
-        this._errorBox.hide();
+        this._descriptionLabel.set_text(_("Press ESC to close"));
+        this._descriptionLabel.remove_style_class_name('error');
         this._entryText.set_text('');
         this._completionBox.hide();
         this._commandError = false;
@@ -432,7 +466,6 @@ __proto__: ModalDialog.ModalDialog.prototype,
         if (this._lockdownSettings.get_boolean(DISABLE_COMMAND_LINE_KEY))
             return;
 
-        ModalDialog.ModalDialog.prototype.open.call(this);
-    },
-};
-Signals.addSignalMethods(RunDialog.prototype);
+        super.open();
+    }
+});
